/*
   xbnfgen version 0.6  Copyright (C) 2007-2014 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/

/* btree.c
   binary tree
*/

#include <stdlib.h>
#include <string.h>

#include "salloc.h"
#include "sutil.h"
#include "btree.h"



NODE * bt_mknode( int namlen, char * nam, int nodsize )
/* ================================================================ */
/* allocates a new node.
   the name is not copied ! just the pointer is kept */
{
NODE * node;

   if( nodsize <= sizeof(NODE) )
      nodsize = sizeof(NODE);
   node = salloc( nodsize );
   node->nodesize = nodsize;
   nstring_assign( node->name, namlen, nam );
   return node;
}


NODE * bt_mktree( char * dummy, int nodsize )
/* ================================================================ */
/* make a new tree by allocating a node with a dummy name.
   dummy must be 0-terminated ! */
{
   return bt_mknode( strlen(dummy), dummy, nodsize );
}


NODE * bt_search( NODE * node, int namlen, char * nam, int * comp )
/* ================================================================ */
/* searches nam in the tree(!) at node
     if found     -> sets *comp  = 0 and returns a pointer to the node
     if not found -> sets *comp != 0 and returns a pointer to
                     the node where the new one must be inserted.
          Insert at node->low  if *comp < 0,
          insert at node->high is *comp > 0    */
{
   *comp = nstrcmp( namlen, nam, nstring(node->name) );
   if( *comp == 0 )
   {
      return node;           /* found */
   }
   else if( *comp < 0 )
   {                         /* nam < node->name ==> go lower  */
      if( node->low == 0 )
         return node;        /* not found */
      else
         return bt_search( node->low , namlen, nam, comp );
   }
   else /* if( *comp > 0 ) */
   {                         /* nam > node->name ==> go higher */
      if( node->high == 0 )
         return node;        /* not found */
      else
         return bt_search( node->high, namlen, nam, comp );
   }
}


NODE * bt_insert( NODE * root, int namlen, char * nam, int * isnew )
/* ================================================================ */
/* inserts nam into the tree
   returns the node containing nam
   *isnew is set to 1 if nam was inserted, 0 if it was already there */
{
int comp;
NODE * node;
NODE * newnode;

   /* search the name: */
   node = bt_search( root, namlen, nam, &comp );
   if( comp == 0 )
   {
      *isnew = 0;
      return node; /* already there */
   }

   /* ok, the name is a new one, we can make a node and insert it */
   newnode = bt_mknode( namlen, nam, root->nodesize );

   /* insert into tree: */
   if( comp < 0 )
      node->low  = newnode;
   else
      node->high = newnode;

   *isnew = 1;
   return newnode;
}


void bt_foreach( NODE * root, NODE * node, BT_PROC proc )
/* ================================================================ */
/* calls function 'proc' for each node of the tree
   depth first walking, i.e. keys are ordered
   the node "root" is ignored (where the dummy element is) */
{
   if( node )
   {
      bt_foreach( root, node->low, proc );
      if( node != root )
         proc( node );
      bt_foreach( root, node->high, proc );
   }
}


