/*
   xbnfgen version 0.8  Copyright (C) 2007-2016 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/

/* scan.c
   general scanner functions */


#include <fcntl.h>
#include <fcntl.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "port.h"
#include "salloc.h"
#include "scan.h"

#define MAXLINELENGTH 17000
#define MAXSYMLEN      2000

#define DUMP_LINE_LEN 10

#define ESC_CHAR '\\'

static _KEY_ emptykeytable[] = { {0} };
static int lineno; // used in scan_dump

SCAN * screadfile( char * fname )
/* ================================================================ */
/* reads the given file and creates the SCAN structure
   (or returns 0 for errors) */
{
int f;
SCAN * inp;
int len, flen;

   /* open the file and determine length */
   f = open( fname, O_RDONLY | O_BINARY );
   if( f == -1 )
   {
      MFATAL( "Cannot open", fname );
      /* should never return, but just to make sure: */
      exit( 2 );
   }

   flen = filelength( f );
   if( flen < 0 )
   {
      MFATAL( "Cannot determine file length of", fname );
      /* should never return, but just to make sure: */
      exit( 2 );
   }
   else if( flen == 0 )
   {
      MFATAL( "File is empty", fname );
      /* should never return, but just to make sure: */
      exit( 2 );
   }

   /* create SCAN structure */
   inp = screadstring( "", -flen );
   if( inp )
   {
      inp->filename   = fname;
      inp->linenocorr = 0;

      /* read the file: */
      len = read( f, inp->src + 1, flen );
      if( len != flen )
      {
         MFATAL( "Error reading", fname );
         /* should never return, but just to make sure: */
         exit( 2 );
      }
      close( f );
   }

   return inp;
}


SCAN * screadstring( char * content, int len )
/* ================================================================ */
// len < 0  : just allocate (-len), no copy
// len == 0 : copy strlen(content)
// len > 0  : copy len
{
SCAN * inp;

   if( !content )
      return 0;

   if( len == 0 )
      len = strlen(content);

   /* make SCAN structure: */
   inp = salloc( sizeof(SCAN) );
   inp->filename = "-";
   inp->filelen = abs(len);

   /* allocate memory: */
   inp->filelen += 2; /* 2 chars buffer: 1 before, 1 after  */
   inp->src = salloc( (inp->filelen)*sizeof(char) );
   inp->sym = salloc( (inp->filelen)*sizeof(_SYMBOL_) );
   inp->att = salloc( (inp->filelen)*sizeof(_ATTRIBUTE_) );

   /* copy content */
   if( len > 0 )
      memcpy( inp->src+1, content, len );

   /* fill before and after: */
   inp->src[0] = '\n';
   inp->sym[0] = BOL,
   inp->att[0].a = 0; // line # (will be incr to 1 due to curbol = 0)
   inp->src[inp->filelen-1] = '\n';
   inp->sym[inp->filelen-1] = EOFSYM,
   inp->att[inp->filelen-1].a = 0;

   /* other elements: */
   inp->comment_begin       = "";
   inp->comment_end         = "";
   inp->comment_line        = "";
   inp->comment_begin_len   = 0;
   inp->comment_end_len     = 0;
   inp->comment_line_len    = 0;
   inp->charsym             = 0;
   inp->optab               = emptykeytable;
   inp->keytab              = emptykeytable;
   inp->keytablen           = 0;
   inp->charlit_sym         = 0;
   inp->charlit_begin_end   = 0;
   inp->stringlit_sym       = 0;
   inp->stringlit_begin_end = 0;
   inp->esc_flags           = 0;
   inp->number_sym          = 0;
   inp->commentlevel        = 0;
   inp->curbol              = 0;
   inp->prevsympos          = 0;
   inp->indent              = 0;

   return inp;
}


int beginofline( SCAN * inp, int pos )
/* ================================================================ */
{
   int i;

   i = pos;
   while( inp->sym[i] != BOL )
   {
      if( inp->sym[i] == NONE )
         i = inp->att[i].a;
      else
         i--;
   }
   return i;
}


static int reallineno( SCAN * inp, int pos )
/* ---------------------------------------------------------------- */
{
   if( pos < 0 )
      return 0;
   return inp->att[beginofline(inp,pos)].a;
}


char * getfileline( SCAN * inp, int pos, int * plineno )
/* ================================================================ */
{
   if( plineno )
      *plineno = reallineno( inp, pos ) + inp->linenocorr;

   return inp->filename;
}


void setfileline( SCAN * inp, int pos, char * filenam, int lineno )
/* ================================================================ */
{
   inp->filename   = filenam;
   inp->linenocorr = lineno - reallineno( inp, pos );
}


int nextsym( SCAN * inp, int pos )
/* ================================================================ */
/* returns the position of the next nontrivial (i.e. non NONE, BOL)
   symbol */
{
int sym, flen;

   flen = inp->filelen;
   for(;;)
   {
      pos++;
      if( pos >= flen )
         return ((unsigned int)(-1))/2; // grosse positive Zahl
      sym = symat(inp,pos);
      if( sym == NXSYM )
      {
         pos = attat(inp,pos).a;
         sym = symat(inp,pos);
         /* double check following, NXSYM might point to NONE ... */
      }
      if( !((sym == NONE) || (sym == BOL)) )
         return nextsymhook( inp, pos );
   }
}


int eatsym( SCAN * inp, int pos, int symbol )
/* ================================================================ */
{
   if( pos > maxpos(inp) )
   {
      MERROR( "<end of file> reached, but expected:", prtermsym( inp, symbol ) );
      return pos;
   }

   if( symat(inp,pos) != symbol )
      MERROR( "expected:", prtermsym( inp, symbol ) );

   pos = nextsym( inp, pos );
   return pos;
}


void putnxsym( SCAN * inp, int pos )
/* ================================================================ */
/* we are at a new symbol position and mark this as next symbol
   at prevsympos+1 (if possible)                              */
{
   if( (pos > (inp->prevsympos+1)) && ((inp->sym[inp->prevsympos+1] == NONE) || (inp->sym[inp->prevsympos+1] == NXSYM)) )
   {
      inp->sym[inp->prevsympos+1]   = NXSYM;
      inp->att[inp->prevsympos+1].a = pos;
   }
   inp->prevsympos = pos;
}


void putsym( SCAN * inp, int pos, int symbol )
/* ================================================================ */
{
   inp->sym[pos] = symbol;
   putnxsym( inp, pos );
}


static char * searchsym( _KEY_ * tab, int symbol )
/* ---------------------------------------------------------------- */
/* looks in table tab[] for symbol, tab[] must end w 0-word
   linear search  */
{
int i;

   if( !tab )
      return 0;

   for( i = 0; tab[i].word; i++ )
      if( tab[i].sym == symbol )
         return tab[i].word;

   return 0;  /* not found */
}


static char * keyop( SCAN * inp, int symbol )
/* ---------------------------------------------------------------- */
/* looks for given symbol in keytab[], optab[], returns string */
{
char * str;

   str = searchsym( inp->keytab, symbol );

   if( !str )
      str = searchsym( inp->optab, symbol );

   return str;
}


static char specialescape( int esc_flags, int ch )
/* ---------------------------------------------------------------- */
{
   switch( ch )
   {
   case '\a': if( esc_flags & escape_a_allowed ) return 'a';  break;
   case '\b': if( esc_flags & escape_b_allowed ) return 'b';  break;
   case '\f': if( esc_flags & escape_f_allowed ) return 'f';  break;
   case '\n': if( esc_flags & escape_n_allowed ) return 'n';  break;
   case '\r': if( esc_flags & escape_r_allowed ) return 'r';  break;
   case '\t': if( esc_flags & escape_t_allowed ) return 't';  break;
   case '\v': if( esc_flags & escape_v_allowed ) return 'v';  break;
   default:;
   }
   return 0;
}


static char * prcharnaked( SCAN * inp, int ch, char * pdelim, int insidestring )
/* ---------------------------------------------------------------- */
// i.e. w/o ''
// *pdelim return delimiter char (begin & end) or 0
// insidestring: == 0 if char only, 8 if octal digit follows, 1 otherwise in a string
{
static char buf[10];
char se;

   *pdelim = inp->charlit_begin_end;
   if( (!insidestring && (ch == inp->charlit_begin_end)) || (insidestring && (ch == inp->stringlit_begin_end))  )
   {
      if( inp->esc_flags & escape_double_start_char )
         buf[0] = ch;
      else
         buf[0] = ESC_CHAR;

      buf[1] = ch;
      buf[2] = 0;
   }
   else if( ch == ESC_CHAR )
   {
      if( inp->esc_flags & escape_hashdec_allowed )
      {
         sprintf( buf, "#%d", ch );
         *pdelim = 0;
      }
      else
      {
         buf[0] = ch;
         buf[1] = ch;
         buf[2] = 0;
      }
   }
   else if( (ch >= 32) && (ch < 127) )
   {
      buf[0] = (char)ch;
      buf[1] = 0;
   }
   else
   {
      if( inp->esc_flags & escape_hashdec_allowed )
      {
         sprintf( buf, "#%d", ch );
         *pdelim = 0;
      }
      else
      {
         se = specialescape( inp->esc_flags, ch );
         if( se )
         {
            sprintf( buf, "%c%c", ESC_CHAR, se );
         }
         else
         {
            if( (ch == 0) && (insidestring < 8) )
            {
               sprintf( buf, "%c0", ESC_CHAR );
            }
            else
            {
               sprintf( buf, "%c%03o", ESC_CHAR, ch & 255 );
            }
         }
      }
   }

   return buf;
}


static char * prchar( SCAN * inp, int ch )
/* ---------------------------------------------------------------- */
{
static char buf[12];
char * naked;
char   delim;

   naked = prcharnaked( inp, ch, &delim, 0 );
   if( delim )
   {
      sprintf( buf, "%c%s%c", delim, naked, delim );
      return buf;
   }
   else
      return naked;
}


static int isoctal( char c )
/* ---------------------------------------------------------------- */
{
   return (c >= '0') && (c <= '7');
}


static char * prstring( SCAN * inp, _STRING_ * str )
/* ---------------------------------------------------------------- */
{
static char buf[MAXLINELENGTH+2];
char * dst;
char   strdelim, chardelim, insidedelims;
char * charstr;
int    i, len, slen;

   strdelim = inp->stringlit_begin_end;
   dst = buf;
   slen = str->val.len;

   if( slen == 0 )
   {
      *dst++ = strdelim;
      insidedelims = 1;
   }
   else
   {
      insidedelims = 0;
   }

   for( i = 0; i < slen; i++ )
   {
      charstr = prcharnaked( inp, str->val.str[i], &chardelim, isoctal(str->val.str[i+1]) ? 8 : 1 );
      len = strlen(charstr);
      if( chardelim )
      {
         if( !insidedelims )
         {
            *dst++ = strdelim;
            insidedelims = 1;
         }
      }
      else
      {
         if( insidedelims )
         {
            *dst++ = strdelim;
            insidedelims = 0;
         }
      }
      strcpy( dst, charstr );
      dst += len;
   }

   if( insidedelims )
      *dst++ = strdelim;
   *dst = 0;
   return buf;
}


static void prreal( char * buf, double val )
/* ---------------------------------------------------------------- */
{
   int a;

   if( val < 0 )
   {
      *buf++ = '-';
      val = -val;
   }
   // xx.0 wenn gannzahlig und nicht zu gross
   if( val > 300 )
   {
      sprintf( buf, "%.16lg", val );
      return;
   }
   a = (int)floor(val);
   if( val == (double)a )
      sprintf( buf, "%d.0", a );
   else
      sprintf( buf, "%.16lg", val );
}


char * prsymat( SCAN * inp, int pos )
/* ================================================================ */
{
   static char buf[40];
   int symbol;
   char * str;

   symbol = symat(inp,pos);

   if( (symbol >= 32) && (symbol <= 127) )
   {
      buf[0] = (char)symbol;
      buf[1] = 0;
      return buf;
   }

   if( (symbol == NONE) || (symbol == BOL) )
      return "";

   if( symbol == EOFSYM )
      return "<end of file>";

   if( symbol == inp->number_sym )
   {
      _NUMBER_ * pn = attat(inp,pos).pn;
      switch( pn->type )
      {
      case INT_NUMBER:
         sprintf( buf, " %" FMT64 "d", pn->intvalue );
         break;
      case REAL_NUMBER:
         *buf = ' ';
         prreal( buf+1, pn->realvalue );
         break;
      default:
         sprintf( buf, "?NumberType%d?", pn->type );
      }
      return buf;
   }

   if( symbol == inp->charlit_sym )
   {
      return prchar( inp, attat(inp,pos).a );
   }

   if( symbol == inp->stringlit_sym )
   {
      return prstring( inp, attat(inp,pos).ps );
   }

   str = keyop( inp, symbol );
   if( str )
      return str;

   return prusersymat( inp, pos );
}


char * prtermsym( SCAN * inp, int symbol )
/* ================================================================ */
{
char * str;

   if( (symbol >= 32) && (symbol <= 127) ) // single char symbols represent themselfs
   {
      static char buf[4];

      buf[0] = '\'';
      buf[1] = (char)symbol;
      buf[2] = '\'';
      buf[3] = 0;
      return buf;
   }

   if( (symbol == NONE) || (symbol == BOL) )
      return "";

   if( symbol == EOFSYM )
      return "<end of file>";

   if( symbol == inp->number_sym )
      return "Number";

   if( symbol == inp->charlit_sym )
      return "Character";

   if( symbol == inp->stringlit_sym )
      return "String";

   str = keyop( inp, symbol );
   if( str )
      return str;

   return prusertermsym( symbol );
}


char * prtermsymset( SCAN * inp, _SET * symset )
/* ================================================================ */
{
static char buf[CARDINAL * 8];
char * bufptr;
char * sym;
int i;

   bufptr = buf;
   for( i = 0; i < CARDINAL; i++ )
   {
      if( set_in( symset, i ) )
      {
         *bufptr++ = ' ';
         sym = prtermsym( inp, i );
         while( *sym )
            *bufptr++ = *sym++;
      }
   }
   *bufptr++ = ' ';
   *bufptr = 0;
   return buf;
}


char * prlineat( SCAN * inp, int pos )
/* ================================================================ */
/* prints the line at the given position (including file name,
   line number and a position marker at the end)
   to a static buffer which is returned */
{
static char buf[MAXLINELENGTH + 400];
char * bufptr;
int i, slen, offset;

   /* goto begin of line: */
   i = beginofline( inp, pos );

   /* filename and line number: */
   sprintf( buf, "%s(%d): ", inp->filename, inp->att[i].a + inp->linenocorr );
   slen = strlen( buf );

   /* char offset of current position: */
   offset = slen + pos - i - 1;

   /* put the line: */
   bufptr = buf + slen;
   i++;
   while( inp->src[i] != '\n' )
      *bufptr++ = inp->src[i++];
   *bufptr++ = '\n';

   /* finally the position marker: */
   for( i = 0; i < offset; i++ )
      *bufptr++ = ' ';
   *bufptr++ = '^';

   *bufptr = 0;
   return buf;
}


void printfromto( FILE * f, SCAN * inp, int from, int to, int braceindent )
/* ================================================================ */
/* prints part of the module from position 'from' to position 'to' */
{
int pos, indent;
_SYMBOL_ prevsym = BOL;

   if( !f )
      return;
   for( pos = from; pos < to; pos++ )
   {
      _SYMBOL_ sym = symat(inp,pos);
      switch( sym )
      {
      case EOFSYM:
         return;

      case NONE:
      case NXSYM:
         break;

      case BOL:
         if( charat(inp,pos) == '\n' )
         {
            indent = inp->indent;
            if( symat(inp,nextsym(inp,pos)) == '}' )
               indent -= braceindent;
            if( indent < 0 )
               indent = 0;
            fprintf( f, "\n%*s", indent, "" );
         }
         break;

      case ',':
      case ';':
         fprintf( f, "%c", symat(inp,pos) );
         break;

      case '(':
      case '[':
         fprintf( f, "%c ", symat(inp,pos) );
         break;

      case '{':
         fprintf( f, "%s%c ", prevsym == BOL ? "" : " ", symat(inp,pos) );
         inp->indent += braceindent;
         break;

      case '}':
         fprintf( f, "%s%c", prevsym == BOL ? "" : " ", symat(inp,pos) );
         inp->indent -= braceindent;
         break;

      default:
         fprintf( f, "%s%s", prevsym == BOL ? "" : " ", prsymat(inp,pos) );
         break;
      }
      if( !((sym==NONE) || (sym==NXSYM)) )
         prevsym = sym;
   }
}


void printmodule( FILE * f, SCAN * inp, int braceindent )
/* ================================================================ */
/* prints the whole module based on the symbols */
{
   printfromto( f, inp, 1, maxpos(inp), braceindent );
}


/* ================================================================ */
void scan_initauto( SCAN * inp,
   char * comment_begin,         // start of comment block
   char * comment_end,           // end of comment block
   char * comment_line,          // start of comment until end of line
   char * charsym,               // 1-char symbols
   _KEY_ *  optab,               // multi-char operators
   _KEY_ *  keytab,              // keywords, alphabetically sorted
   _SYMBOL_ charlit_sym,         // if != 0: automatic char literal handling
   char   charlit_begin_end,     // e.g. '
   _SYMBOL_ stringlit_sym,       // if != 0: automatic string literal handling
   char   stringlit_begin_end,   // e.g. "
   int    feature_flags,         // see feature_... below
   int    esc_flags,             // flags for char and string literals // escape_... below
   _SYMBOL_ number_sym           // if != 0: automatic number handling
   )
{
   int i;

   if( !inp )
      return;

   inp->charsym             = charsym;
   inp->charlit_sym         = charlit_sym;
   inp->charlit_begin_end   = charlit_begin_end;
   inp->stringlit_sym       = stringlit_sym;
   inp->stringlit_begin_end = stringlit_begin_end;
   inp->feature_flags       = feature_flags;
   inp->esc_flags           = esc_flags;
   inp->number_sym          = number_sym;

   inp->keytablen = 0;
   if( keytab )
   {
      inp->keytab = keytab;
      for( i = 0; keytab[i].word; i++ )
      {
         inp->keytab[i].wordlen = strlen(keytab[i].word);
         inp->keytablen++;
      }
   }
   else
   {
      inp->keytab = emptykeytable;
   }

   inp->optab = optab;
   if( optab )
   {
      for( i = 0; optab[i].word; i++ )
         inp->optab[i].wordlen = strlen(optab[i].word);
   }

   if( comment_begin )
   {
      inp->comment_begin     = comment_begin;
      inp->comment_begin_len = strlen(comment_begin);
   }
   else
   {
      inp->comment_begin     = "";
      inp->comment_begin_len = 0;
   }

   if( comment_end )
   {
      inp->comment_end     = comment_end;
      inp->comment_end_len = strlen(comment_end);
   }
   else
   {
      inp->comment_end     = "";
      inp->comment_end_len = 0;
   }

   if( comment_line )
   {
      inp->comment_line     = comment_line;
      inp->comment_line_len = strlen(comment_line);
   }
   else
   {
      inp->comment_line     = "";
      inp->comment_line_len = 0;
   }
}


int posinc( SCAN * inp, int pos, int inc )
/* ================================================================ */
{
   int p;
   for( p = pos+1; p <= pos+inc; p++ )
      attat(inp,p).a = inp->curbol;
   return pos+inc;
}


static int isthisstring( SCAN * inp, int pos, int slen, char * sptr )
/* ---------------------------------------------------------------- */
/* Is string s (sptr, slen) at pos * */
{
   if( slen <= 0 )
      return 0;

   return nstrcmp( slen, sptr, slen, strptrat(inp,pos) ) == 0;
}


int putnewline( SCAN * inp, int pos )
/* ================================================================ */
{
   putsym(inp,pos,BOL);
   attat(inp,pos).a = attat(inp,inp->curbol).a + 1; // line #
   inp->curbol = pos;
   return posinc( inp, pos, 1 );
}


int scan_comment( SCAN * inp, int pos )
/* ================================================================ */
/* returns 1st position (starting at pos) outside a comment */
{
   int maxp = maxpos(inp);
   for(;;)
   {
      if( pos > maxp )
         return maxp+1;
      while( inp->commentlevel > 0 )
      {
         if( pos > maxp )
            return maxp+1;
         if( isthisstring( inp, pos, inp->comment_end_len, inp->comment_end ) )
         {
            inp->commentlevel--;
            pos = posinc( inp, pos, inp->comment_end_len );
         }
         else if( isthisstring( inp, pos, inp->comment_begin_len, inp->comment_begin ) )
         {
            if( !(inp->feature_flags & feature_comment_not_nested) )
               inp->commentlevel++;
            pos = posinc( inp, pos, inp->comment_begin_len );
         }
         else if( charat(inp,pos) == '\n' )
         {
            pos = putnewline( inp, pos );
         }
         else
            pos = posinc( inp, pos, 1 );
      }
      // commentlevel is 0 here

      if( isthisstring( inp, pos, inp->comment_begin_len, inp->comment_begin ) )
      {
         inp->commentlevel++;
         pos = posinc( inp, pos, inp->comment_begin_len );
      }
      else if( isthisstring( inp, pos, inp->comment_line_len, inp->comment_line ) )
      {
         while( charat(inp,pos) != '\n' )   // comment until EOL
         {
            if( pos > maxp )
               return maxp+1;
            attat(inp,pos).a = inp->curbol;
            pos = posinc( inp, pos, 1 );
         }
      }
      else if( isthisstring( inp, pos, inp->comment_end_len, inp->comment_end ) )
      {
         MERROR( "unmatched end of comment", 0 );
         pos = posinc( inp, pos, inp->comment_end_len );
      }
      else
      {
         return pos;
      }
   }
}

static int gethex( SCAN * inp, int pos, int64 * pvalue )
/* ---------------------------------------------------------------- */
{
static char dec[] = "0123456789-------:;<=>?--------------------------:;<=>?";
char c;
int n;
int64 nu = 0;

   for(;;)
   {
      c = charat(inp,pos);
      if( c < '0' || c > 'f' )
         break;    /* not a hex digit --> end of hex number */
      n = dec[c-'0'] - '0';
      if( n < 0 )
         break;    /* not a hex digit --> end of hex number */
      nu = nu * 16 + n;
      pos = posinc(inp,pos,1);
   }
   *pvalue = nu;
   return pos;
}


static int getint( SCAN * inp, int pos, int base, int64 * pvalue, int * pdigits )
/* ---------------------------------------------------------------- */
/* base <= 10
   wenn pdigits != 0, dann
      gibt *pdigits die Anzahl der Ziffern zurueck
      ist *pdigits > 0 beim Aufruf von getint(),
         dann ist es die maximale Anzahl von Ziffern */
{
char c;
int64 nu = 0;
int64 nunew = 0;
int maxdigits = 0x7FFFFFFF;

   if( pdigits )
   {
      if( *pdigits > 0 )
         maxdigits = *pdigits;
      *pdigits = 0;
   }
   for(;;)
   {
      c = charat(inp,pos);
      if( c < '0' || c >= ('0' + base) )
         break;    /* not a digit --> end of number */
      if( nunew >= 0 )
         nunew = nu * base + c - '0';
      pos = posinc(inp,pos,1);
      if( nunew >= 0 )
      {
         nu = nunew;
         if( pdigits )
         {
            (*pdigits)++;
            if( *pdigits >= maxdigits )
               break;
         }
      }
   }
   *pvalue = nu;
   return pos;
}


static int escape( SCAN * inp, int pos, int64 * pvalue )
/* ---------------------------------------------------------------- */
/* scan 1 char escape */
{
   int maxoctaldigits = 3;
   char ch = charat(inp,pos);

   switch( ch )
   {
   case 'a': if( inp->esc_flags & escape_a_allowed ) { *pvalue = '\a'; return pos+1; }   break;
   case 'b': if( inp->esc_flags & escape_b_allowed ) { *pvalue = '\b'; return pos+1; }   break;
   case 'f': if( inp->esc_flags & escape_f_allowed ) { *pvalue = '\f'; return pos+1; }   break;
   case 'n': if( inp->esc_flags & escape_n_allowed ) { *pvalue = '\n'; return pos+1; }   break;
   case 'r': if( inp->esc_flags & escape_r_allowed ) { *pvalue = '\r'; return pos+1; }   break;
   case 't': if( inp->esc_flags & escape_t_allowed ) { *pvalue = '\t'; return pos+1; }   break;
   case 'v': if( inp->esc_flags & escape_v_allowed ) { *pvalue = '\v'; return pos+1; }   break;

   case 'x': if( inp->esc_flags & escape_x_allowed ) return gethex( inp, pos+1, pvalue );break;
   case 'X': if( inp->esc_flags & escape_X_allowed ) return gethex( inp, pos+1, pvalue );break;

   case '0': case '1': case '2': case '3':
   case '4': case '5': case '6': case '7':
      if( inp->esc_flags & escape_octal_allowed ) return getint( inp, pos, 8, pvalue, &maxoctaldigits );
      break;

   default:;
   }
   *pvalue = ch;
   return pos+1;
}


static int character( SCAN * inp, int pos )
/* ---------------------------------------------------------------- */
/* scans a character literal starting at inp->src[pos],
   attribut: att[pos] = char value
 */
{
int j, charfound;
int64 val;
char ch;
int maxp = maxpos(inp);

   _symat(inp,pos) = inp->charlit_sym;
   j = pos;
   val = 0;
   charfound = 0;
   for(;;)
   {
      if( j > maxp )
      {
         MERROR_POS( j-1, "end of file in character literal", 0 );
         break;
      }
      ch = charat(inp,j);
      if( ch == '\n' )
      {
         MERROR_POS( j-1, "end of line in character literal", 0 );
         j++;
         break;
      }
      if( ch == '\\' )
      {
         j = escape( inp, j+1, &val );
         charfound++;
         continue;
      }
      attat(inp,j).a = inp->curbol;
      j++;
      if( ch == inp->charlit_begin_end )
      {
         if( (inp->esc_flags & escape_double_start_char) && (charat(inp,j) == ch) )
         {
            val = ch;
            charfound++;
            continue;
         }
         else
            break; // end of char literal
      }
      val = ch;
      charfound++;
   }
   attat(inp,pos).a = (int)val;
   putnxsym( inp, pos );
   if( charfound == 0 )
      MERROR_POS( j, "character missing", 0 );
   else if( charfound > 1 )
      MERROR_POS( j-1, "more than one character", 0 );
   return j;
}


static void stringtextlen( SCAN * inp, int pos, int * textlen, char * esc )
/* ---------------------------------------------------------------- */
{
int j;
char ch;
int maxp = maxpos(inp);

   *esc = 0;
   j = pos;
   for(;;)
   {
      if( j == maxp )
         break;
      ch = charat(inp,j);
      if( ch == '\n' )
         break;
      if( (ch == '\\') && (inp->esc_flags & (escape_std_c | escape_X_allowed)) )
      {
         *esc = 1;
         j = posinc(inp,j,1);
      }
      else if( (ch == inp->stringlit_begin_end) )
      {
         if( (inp->esc_flags & escape_double_start_char) && (charat(inp,j+1) == ch) )
         {
            *esc = 1;
            j = posinc(inp,j,1);
         }
         else if( (inp->esc_flags & escape_hashdec_allowed) && (charat(inp,j+1) == '#') )
         {
            *esc = 1;
            j = posinc(inp,j,1);
          hash: // charat(inp,j) == '#'
            j = posinc(inp,j,3);
            ch = charat(inp,j);
            if( ch == '#' )
               goto hash;
            else if( ch == inp->stringlit_begin_end )
               ;
            else
               break;
         }
         else
            break;
      }
      j = posinc(inp,j,1);
   }
   *textlen = j - pos;
}


static int string( SCAN * inp, int pos )
/* ---------------------------------------------------------------- */
/* scans a string literal starting at inp->src[pos],
   attribut: att[pos] = pointer to STRING
 */
{
int j, textlen;
int64 val;
char ch, esc;
_STRING_ * pstring;
char * dest;
int maxp = maxpos(inp);

   // wir wollen zuerst die Laenge im Text wissen, und ob esc-Zeichen enthalten sind
   stringtextlen( inp, pos, &textlen, &esc );
   /* wir muessen in jedem Fall die STRING-struct allozieren.
      Falls escape-Zeichen vorhanden sind, dann auch noch Platz fuer den
      escape-ausgeloesten String. */
   pstring = salloc( sizeof(_STRING_) );
   pstring->txt.len  = textlen;
   pstring->txt.str  = strptrat(inp,pos);
   _symat(inp,pos)   = inp->stringlit_sym;
   attat(inp,pos).ps = pstring;
   putnxsym( inp, pos );
   if( !esc )
   {
      pstring->val = pstring->txt;
      attat(inp,pos+textlen+1).a = inp->curbol;
      return pos + textlen + 1;
   }
   else
   {
      pstring->val.str = (char *)salloc(textlen+1); // laeanger kann es nicht werden (+1 wg. 0 am Ende)

      // hier nun den string einfuellen, dabei escapes aufloesen
      dest = pstring->val.str;
      j = pos;
      for(;;)
      {
         if( j > maxp )
         {
            MERROR_POS( j-1, "end of file in string literal", 0 );
            break;
         }
         ch = charat(inp,j);
         if( ch == '\n' )
         {
            MERROR_POS( j-1, "end of line in string literal", 0 );
            j++;
            break;
         }
         if( ch == '\\' )
         {
            j = escape( inp, j+1, &val );
            *dest++ = (char)val;
            continue;
         }
         if( (ch == inp->stringlit_begin_end) )
         {
            if( (inp->esc_flags & escape_double_start_char) && (charat(inp,j+1) == ch) )
            {
               *dest++ = ch;
               j++;
               continue;
            }
            else
            {
               j++;
               break; // end of string literal
            }
         }
         *dest++ = ch; // ordinary char
         j++;
      }
      pstring->val.len = dest - pstring->val.str;
      *dest = 0; // schadet nicht...
   }
   return j;
}


static _KEY_ * op1stchar( _KEY_ * optab, char ch )
/* ---------------------------------------------------------------- */
/* returns 1st entry of optab which begin w ch */
{
   if( !optab )
      return 0;

   while( optab->word )
   {
      if( optab->word[0] == ch )
         return optab;
      optab++;
   }
   return 0;
}


static int multicharoperator( SCAN * inp, int pos )
/* ---------------------------------------------------------------- */
/* returns new pos or 0 */
{
   char ch = charat(inp,pos);
   _KEY_ * operator;
   operator = op1stchar( inp->optab, ch );
   while( operator )
   {
      if( nstrcmp(operator->wordlen,operator->word,operator->wordlen,strptrat(inp,pos)) == 0 )
      {
         putsym(inp,pos,operator->sym);
         attat(inp,pos).a = operator->wordlen;
         return posinc( inp, pos, operator->wordlen );
      }
      operator = op1stchar( operator+1, ch );
   }
   return 0;
}


static int keycmp( const void * s , const void * k )
/* ---------------------------------------------------------------- */
                 /* s -> lstring    k -> keytab entry */
{
   /* case sensitive */
   return nstrcmp( nstring(*(NSTRING *)s), ((_KEY_ *)k)->wordlen, ((_KEY_ *)k)->word );
}


static int keycmpi( const void * s , const void * k )
/* ---------------------------------------------------------------- */
                 /* s -> lstring    k -> keytab entry */
{
   /* case INsensitive */
   return nstricmp( nstring(*(NSTRING *)s), ((_KEY_ *)k)->wordlen, ((_KEY_ *)k)->word );
}


static int keyorident( SCAN * inp, int pos )
/* ---------------------------------------------------------------- */
/* scans a word starting at inp->src[pos],
   fills inp->sym/n i ... until end of word
   if in keytab -> keyword, else IDENT
   returns index at end of word             */
{
int j;
NSTRING lstr;
_KEY_ * k;
char ch;

   /* determine the length of the word: */
   j = pos;
   for(;;)
   {
      attat(inp,j).a = inp->curbol;
      ch = charat(inp,j);
      if( ch < '0' || ch > 'z' )
         break;
      if( ch > '9' && ch < 'A' )
         break;
      if( ch > 'Z' && ch < 'a' && ch != '_' )
         break;
      j++;
   }
   lstr.len = j - pos;
   if( lstr.len == 0 )
   {
      MERROR_POS( j, "invalid character",
           (unsigned char)ch < ' ' ? strf("(0x%02X)",ch & 255) : strf("'%c'=0x%02X",ch,ch & 255) );
      j++;
      lstr.len = 1;
   }
   lstr.str = strptrat(inp,pos);
   strlenat(inp,pos) = lstr.len;

   /* search in key word table: */
   k = bsearch( &lstr, inp->keytab, inp->keytablen, sizeof(_KEY_), inp->feature_flags & feature_case_sensitive ? keycmp : keycmpi );

   _symat(inp,pos) = k ? k->sym : wordsymbol(inp,pos);

   // special case: if sym == stringlit_sym we need to use .ps instead of .a
   if( symat(inp,pos) == inp->stringlit_sym )
   {
      _STRING_ * pstr = salloc(sizeof(_STRING_));
      pstr->txt = lstr;
      pstr->val = lstr;
      attat(inp,pos).ps = pstr;
   }

   putnxsym( inp, pos );
   return j;
}


static _NUMBER_ * newint( int64 value )
/* ---------------------------------------------------------------- */
{
   _NUMBER_ * num;

   num = salloc( sizeof(_NUMBER_) );
   num->type = INT_NUMBER;
   num->intvalue  = value;
   num->realvalue = (double)value;

   return num;
}


static _NUMBER_ * newreal( int64 intval, int64 fracval, int fraclen, int64 expval )
/* ---------------------------------------------------------------- */
{
   _NUMBER_ * num;

   num = salloc( sizeof(_NUMBER_) );
   num->type = REAL_NUMBER;
   num->intvalue  = intval;
   num->realvalue = ((double)intval + (double)fracval / pow(10,fraclen) ) * pow(10,(double)expval);

   return num;
}


static int number( SCAN * inp, int pos )
/* ---------------------------------------------------------------- */
/* scans decimal number starting at inp->src[pos],
   returns pos at end of number
    digit { digit }  (integer)
    digit { digit } . { digit } E [ < + | - > ] digit { digit } (real)
 */
{
int j, jj, basedigits, baselen, fracdigits, intbase;
int64 intval, fracval, expval;
char nch;

   // may be hex number ?
   if( inp->feature_flags & feature_hex_numbers )
   {
      if( (charat(inp,pos)=='0') && ((charat(inp,pos+1) & 0x5F) =='X') && (inp->esc_flags & escape_x_allowed) )
         return lit_hexint( inp, pos+2 );
   }

   _symat(inp,pos) = inp->number_sym;
   putnxsym( inp, pos );

   jj = pos;
   /* now, in any case the number starts with an integer */

   if( (inp->feature_flags & feature_octal_numbers) && (charat(inp,pos)=='0') )    /* but may be octal ... */
      intbase = 8;
   else
      intbase = 10;
   basedigits = 0;
   j = getint( inp, jj, intbase, &intval, &basedigits );
   baselen = j - jj;
   if( intbase == 10 )
   {
      switch( charat(inp,j) )
      {
      case '.':
         nch = charat(inp,j+1);
         if( (nch >= '0') && (nch <= '9') )
         {  // real, fractional part following
            fracdigits = 0;
            j = getint( inp, j+1, 10, &fracval, &fracdigits );
            nch = charat(inp,j);
            expval = 0;
            if( (nch == 'E') || (nch == 'e') )
               j = getint( inp, j+1, 10, &expval, 0 );
            attat(inp,pos).pn = newreal( intval, fracval, fracdigits, expval + baselen-basedigits );
            return j;
         }
         else if( (nch == 'E') || (nch == 'e') )
         {  // real,  exponent following
            j = getint( inp, j+2, 10, &expval, 0 );
            attat(inp,pos).pn = newreal( intval, 0, 0, expval + baselen-basedigits );
            return j;
         }
         else
         {  // integer
            attat(inp,pos).pn = newint( intval );
            return j;
         }

      case 'E':  case 'e':
         // real, exponent following
         j = getint( inp, j+1, 10, &expval, 0 );
         attat(inp,pos).pn = newreal( intval, 0, 0, expval + baselen-basedigits );
         return j;

      default:;   // integer
      }
   }
   // integer
   attat(inp,pos).pn = newint( intval );
   return j;
}


int lit_hexint( SCAN * inp, int pos )
/* ================================================================ */
{
   int j;
   int64 value;

   j = gethex( inp, pos, &value );
   if( j > pos )
   {
      _symat(inp,pos)   = inp->number_sym;
      attat(inp,pos).pn = newint( value );
      putnxsym(inp,pos);
   }
   return j;
}


int scan_auto( SCAN * inp, int pos )
/* ================================================================ */
{
   int  newpos;
   char ch;
   int maxp = maxpos(inp);
   for(;;)
   {
      pos = scan_comment( inp, pos );
      if( pos > maxp )
         return maxp+1;
      ch = charat(inp,pos);
      if( (ch == '\r') || (ch == '\t') )
      {
         charat(inp,pos) = ' ';
         pos = posinc( inp, pos, 1 );
      }
      else if( ch == ' ' )
      {
         pos = posinc( inp, pos, 1 );
      }
      else if( ch == '\n' )
      {
         pos = putnewline( inp, pos );
      }
      else if( inp->charlit_sym && (ch == inp->charlit_begin_end) )
      {
         return character( inp, pos+1 );
      }
      else if( inp->stringlit_sym && (ch == inp->stringlit_begin_end) )
      {
         return string( inp, pos+1 );
      }
      else if( inp->number_sym && (ch >= '0') && (ch <= '9') )
      {
         return number( inp, pos );
      }
      else if( (newpos = multicharoperator(inp,pos)) != 0 )
      {
         return newpos;
      }
      else if( inp->charsym && strchr(inp->charsym,ch) )
      {
         putsym(inp,pos,ch);
         return posinc( inp, pos, 1 );
      }
      else if( inp->charlit_sym && (inp->esc_flags & escape_hashdec_allowed) && (ch == '#') )
      {
         int64 val;
         newpos = gethex( inp, pos+1, &val );
         putsym(inp,pos,inp->charlit_sym);
         return newpos;
      }
      else if( ((ch >= 'A') && (ch <= 'Z')) || ((ch >= 'a') && (ch <= 'z')) || (ch == '_') )
      {
         return keyorident( inp, pos );
      }
      else
         return pos; // user symbol
   }
}


static void tab_dump( FILE * f, _KEY_ * tab )
/* ---------------------------------------------------------------- */
/* tab[] must end w 0-word  */
{
int i;

   if( !tab )
      return;

   for( i = 0; tab[i].word; i++ )
      fprintf( f, "   %3d: '%s'\n", tab[i].sym, tab[i].word );
}


void src_dump( FILE * f, char * src, int len )
/* ---------------------------------------------------------------- */
{
   int i;
   char c;

   for( i = 0; i < len; i++ )
   {
      c = src[i];
      if((c >= ' ') && (c <= '~'))
         fprintf( f, " '%c' ", c );
      else
         fprintf( f, " x%02X ", c & 255 );
   }
   fprintf( f, "\n" );
}


void sym_dump( FILE * f, _SYMBOL_ * sym, int len )
/* ---------------------------------------------------------------- */
{
   int i;

   for( i = 0; i < len; i++ )
   {
      fprintf( f, " %3d ", sym[i] );
   }
   fprintf( f, "\n" );
}


static char * strdump( char * src, int len )
/* ---------------------------------------------------------------- */
{
   static char dst[5];
   int i, j;

   for( i = 0; i < (sizeof(dst)); i++ )
      dst[i] = 0;

   if( len > 4 )
      len = 4;

   if( len < 4 )
   {
      dst[0] = '"';
      j = 1;
   }
   else
      j = 0;

   for( i = 0; i < len; i++, j++ )
   {
      dst[j] = src[i];
      if( dst[j] < ' ' )
         dst[j] = ' ';
   }
   return dst;
}


void att_dump( FILE * f, SCAN * inp, _SYMBOL_ * sym, _ATTRIBUTE_ * att, int startpos, int len )
/* ---------------------------------------------------------------- */
{
   int  i;

   for( i = 0; i < len; i++ )
   {
      if( sym[i] == BOL )
      {
         lineno = att[i].a;
         fprintf( f, "%4d#", lineno % 10000 );
      }
      else if( sym[i] == NXSYM )
      {
         fprintf( f, "%4d>", att[i].a % 10000 ); // index of next symbol
      }
      else if( sym[i] == EOFSYM )
      {
         fprintf( f, " EOF " );
      }
      else if( inp->charlit_sym && (sym[i] == inp->charlit_sym) )
      {
         fprintf( f, " x%02X ", att[i].a & 255 );
      }
      else if( inp->stringlit_sym && (sym[i] == inp->stringlit_sym) )
      {
         NSTRING nstr = att[i].ps->val;
         fprintf( f, "%s\"", strdump(nstr.str,nstr.len) );
      }
      else if( inp->number_sym && (sym[i] == inp->number_sym) )
      {
         _NUMBER_ * pn = att[i].pn;
         switch( pn->type )
         {
         case INT_NUMBER:  fprintf( f, "%4" FMT64 "d ", pn->intvalue ); break;
         case REAL_NUMBER: fprintf( f, "%.3lg ", pn->realvalue );       break;
         default:          fprintf( f, "?N%2d ", pn->type );
         }
      }
      else if( (sym[i] == wordsymbol(inp,startpos+i)) || (keyop(inp,sym[i]) != 0) )
      {
         fprintf( f, "L%3d ", att[i].a ); // strlen
      }
      else
      {   // --> BOL
         fprintf( f, "%4d<", att[i].a % 10000 ); // index of begin of line
      }
   }
   fprintf( f, "\n" );
}


void scan_dump( FILE * f, SCAN * inp )
/* ================================================================ */
{
   int i, len;

   fprintf( f, " scan_dump begin *****************************************************\n" );
   fprintf( f, "filename = '%s'\n", inp->filename  );
   fprintf( f, "filelen  = %ld\n" , inp->filelen-2 );
   fprintf( f, "comment_begin       = %s\n", inp->comment_begin );
   fprintf( f, "comment_end         = %s\n", inp->comment_end   );
   fprintf( f, "comment_line        = %s\n", inp->comment_line  );
   fprintf( f, "charsym             = %s\n", inp->charsym );
   fprintf( f, "optab:\n"  ); tab_dump( f, inp->optab  );
   fprintf( f, "keytab:\n" ); tab_dump( f, inp->keytab );
   fprintf( f, "charlit_sym         = %3d\n" , inp->charlit_sym         );
   fprintf( f, "charlit_begin_end   = %3c\n" , inp->charlit_begin_end   ? inp->charlit_begin_end   : ' ' );
   fprintf( f, "stringlit_sym       = %3d\n" , inp->stringlit_sym       );
   fprintf( f, "stringlit_begin_end = %3c\n" , inp->stringlit_begin_end ? inp->stringlit_begin_end : ' ' );
   fprintf( f, "feature_flags       = %03X\n", inp->feature_flags  );
   fprintf( f, "esc_flags           = %03X\n", inp->esc_flags  );
   fprintf( f, "number_sym          = %3d\n" , inp->number_sym );

   fprintf( f, "\n line      pos  src/sym/att:" );
   lineno = 1;
   for( i = 0; i < inp->filelen; i += DUMP_LINE_LEN )
   {
      len = DUMP_LINE_LEN;
      if( i + len > inp->filelen )
         len = inp->filelen - i;
      fprintf( f, "\n%5d: %7d ", lineno, i ); src_dump( f, inp->src+i, len );
      fprintf( f, "              "         ); sym_dump( f, inp->sym+i, len );
      fprintf( f, "              "         ); att_dump( f, inp, inp->sym+i, inp->att+i, i, len );
   }

   fprintf( f, " scan_dump end   *****************************************************\n" );
}
