#ifndef _SCAN_H
#define _SCAN_H
/*
   xbnfgen version 0.6  Copyright (C) 2007-2014 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/

/* scan.h
   general scanner definitions

  This is the C interface for a 'new type scanner' - which means
   reads the input file at once
   keeps token information in arrays "parallel" to the input file
   doesn't copy strings from the source
   uses original strings 'in place'
   does not use zero-terminated strings, but strings with explicite length
   may collect more information then a traditional scanner
  advantages:
   faster
   simpler
   makes parser simpler (e.g. look-ahead is trivial)
  disadvantage:
   needs much more memory.

  Using xbnfgen the creation of a scanner is very easy, because most of the
  work is done automatically. see sample pl0*.*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "int64.h"
#include "nstring.h"
#include "set.h"
#include "sutil.h"

/* some generic symbols.
   Note that we assume here that not more then 256 symbols are
   needed (see below). Otherwise CARDINAL in set.h must be increased ! */
#define NONE         0       /* inside a token, comment, white space   */
#define BOL          1       /* begin of line                          */
#define NXSYM        2       /* pointer to next symbol                 */
#define EOFSYM     255       /* end of file                            */

/* symbol assumptions:
    max 256 symbols, all symbol values in the range 0..255, NONE == 0
    single char symbols : symbol value == ASCII code of that char
*/


/* The following type SCAN contains the scanned file including the
   complete source code for each character symbol and attribute.
   src, sym and att are arrays that are allocated by scan().
   All arrays have the same length: filelen+2 (there are additional
   linefeed characters added before and after the file contents).
   sym[i] and att[i] belong to src[i]. For tokens that are longer than
   1 char the symbol is associated with the first char of that token.
   The meaning of the attribute depends on the symbol:
    symbol                                   attribute
     BOL                                      line number            .a
     identifier, keyword, multichar-operator  length of the string   .a
     string                                   pointer to STRING      .ps
     number                                   pointer to NUMBER      .pn
     char                                     char value             .a
     other                                    index of current BOL   .a
*/

typedef unsigned char _SYMBOL_;

typedef struct
{
   char * word; // 0 = end of table
   _SYMBOL_ sym;
   int wordlen;
} _KEY_;

typedef struct
{
  unsigned char type; // XXX_NUMBER below
  int64  intvalue;
  double realvalue;
} _NUMBER_;
#define INT_NUMBER  1
#define REAL_NUMBER 2

typedef struct
{
  NSTRING txt; // so wie im Programmtext, zeigt auf src[]
  NSTRING val; // der von escape-Folgen befreite string
} _STRING_;

typedef union
{
  int        a;
  _NUMBER_ * pn;     // iff sym == number_sym
  _STRING_ * ps;     // iff sym == stringlit_sym
} _ATTRIBUTE_;

typedef struct
{
   char        * filename;
   long          filelen;
   int           linenocorr;
   char        * src;
   _SYMBOL_    * sym;
   _ATTRIBUTE_ * att;
   char        * comment_begin;       // start of comment block
   char        * comment_end;         // end of comment block
   char        * comment_line;        // start of comment until end of line
   unsigned char comment_begin_len;
   unsigned char comment_end_len;
   unsigned char comment_line_len;
   char        * charsym;             // 1-char symbols
   _KEY_       * optab;               // multi-char operators
   _KEY_       * keytab;              // keywords, alphabetically sorted
   int           keytablen;
   _SYMBOL_      charlit_sym;         // if != 0: automatic char literal handling
   char          charlit_begin_end;   // e.g. '
   _SYMBOL_      stringlit_sym;       // if != 0: automatic string literal handling
   char          stringlit_begin_end; // e.g. "
   int           feature_flags;       // see feature_... below
   int           esc_flags;           // flags for char and string literals // escape_... below
   _SYMBOL_      number_sym;          // if != 0: automatic number handling
   /* used during scan: */
   int           commentlevel;
   int           curbol;              /* where the current line begins */
   int           prevsympos;          /* where the previous symbol was - to place NXSYM */
   int           indent;              /* for printfromto() */
} SCAN;


/* *** an error message function must be implemented somewhere ! -- sample: conmsg.c */
void msg( int kind, SCAN * inp, int pos, char * txt, char * specific, char * srcfile, int srcline );
/* a message to the user according to the kind constants below.
   inp/pos indicate error position, can be 0
   txt is the message, specific : some specific string like symbol name
   srcfile/srcline for __FILE__, __LINE__ */
/* msg() should count errors and warnings ... : */
void geterrcnt( int * errors, int * warnings );
/* messages go to max 3 files: */
void setmsgfiles( FILE * out1, FILE * out2, FILE * out3 );


/* msg kind constants: */
#define MSG_WARNING 0   /* points to something which may contain a bug  */
#define MSG_ERROR   1   /* points to something that contains a bug      */
#define MSG_FATAL   2   /* cannot continue, the program ends (in msg()) */
#define MSG_DEBUG   3   /* can be anything                              */


/* macros that use msg:, inp & pos must be defined */
#define MWARNING(txt,specific)       msg(MSG_WARNING,inp,pos,txt,specific,__FILE__,__LINE__)
#define MERROR(txt,specific)         msg(MSG_ERROR,  inp,pos,txt,specific,__FILE__,__LINE__)
#define MERROR_POS(pos,txt,specific) msg(MSG_ERROR,  inp,pos,txt,specific,__FILE__,__LINE__)
#define MFATAL_POS(pos,txt,specific) msg(MSG_FATAL,  inp,pos,txt,specific,__FILE__,__LINE__)
#define MFATAL(txt,specific)         msg(MSG_FATAL,  0,  0,  txt,specific,__FILE__,__LINE__)
#define MDEBUG(txt,specific)         msg(MSG_DEBUG,  inp,pos,txt,specific,__FILE__,__LINE__)


/* *** functions implemented in XYZscan.c : *** */

SCAN * scan( SCAN * inp );
/* scans the given file or returns 0 for fatal errors
   inp is the result of screadfile  -- sample: pl0scan.c */

#ifdef SCAN_NEXTSYM_HOOK
int nextsymhook( SCAN * inp, int pos );
/* call-back function. called for each symbol position.
   normally you should just return pos. */
#else
static inline int nextsymhook( SCAN * inp, int pos ) { return pos; }
#endif

int wordsymbol( SCAN * inp, int pos );
/* call-back function. the scanner found a symbol which is a word but not a keyword.
   give the symbol value (normally IDENT) */

char * prusersymat( SCAN * inp, int pos );
/* prints the symbol at the given position
   to a static buffer which is returned -- only for the 'user symbols' */

char * prusertermsym( int symbol );
/* prints the (terminal) symbol name in readable form
   for errors mesaages etc.
   to a static buffer which is returned -- only for the 'user symbols' */

/* user symbols: all symbols not handled automatically.
   The following symbols are handled automatiaclly when using xbnfgen and
   are therefore no user symbols:
    all keywords
    all operators explicitely given in the xbnf language definition
    numbers (integer and real)
    char and string literals
   User symbols:
    identifiers ("half user symbol" : automatic scan, but needs wordsymbol(), prusersymat(), prusertermsym() )
    user defined operators (if the language permits this)
    special number formats...
*/

/* *** functions implemented in scan.c and some generic macros: *** */

typedef char * (* PRTERMSYM)( int symbol );

SCAN * screadfile( char * fname ); /* 1st function to call ! */
/* reads the given file and creates the SCAN structure
   (or returns 0 for errors) */

SCAN * screadstring( char * content, int len ); // alternative to screadfile()
// len < 0  : just allocate (-len), no copy
// len == 0 : copy strlen(content)
// len > 0  : copy len

char * getfileline( SCAN * inp, int pos, int * plineno ); // get current file/line
void   setfileline( SCAN * inp, int pos, char * filenam, int lineno ); // set current file/line for error messages

int beginofline( SCAN * inp, int pos );
/* position of begin of line */

int posinc( SCAN * inp, int pos, int inc );
/* just increment the current position (e.g. to skip something) */

int nextsym( SCAN * inp, int pos );
/* returns the position of the next nontrivial (i.e. non NONE, BOL)
   symbol */

int eatsym( SCAN * inp, int pos, int symbol );
/* expects the current symbol to be symbol, and 'eats' it, i.e.
   returns the position of the next symbol.
   otherwise an error message is generated */

void putnxsym( SCAN * inp, int pos );
/* we are at a new symbol position and mark this as next symbol
   at prevsympos+1 (if possible) */

void putsym( SCAN * inp, int pos, int symbol );
/* store symbol in sym, calls putnxsym */

int putnewline( SCAN * inp, int pos );

char * prsymat( SCAN * inp, int pos );
/* prints the symbol at the given position
   to a static buffer which is returned */

char * prtermsym( SCAN * inp, int symbol );
/* prints the (terminal) symbol value in readable form
   for errors messages etc.
   to a static buffer which is returned */

char * prtermsymset( SCAN * inp, _SET * symset );
/* prints the given set of (terminal) symbol values in readable form
   to a static buffer which is returned. */

char * prlineat( SCAN * inp, int pos );
/* prints the line at the given position (including file name,
   line number and a position marker at the end)
   to a static buffer which is returned */

/* file output based in scanned symbols: */
void printfromto( FILE * f, SCAN * inp, int from, int to, int braceindent );
/* prints everything from position 'from' to position 'to' to the file given,
   mit brace indentation as given (0 for none) */

void printmodule( FILE * f, SCAN * inp, int braceindent );
/* prints EVERYTHING to the file given */


/* helper macros:  inp must be of type SCAN */
#define maxpos( inp )     ((inp)->filelen-2)
/* max possible position */

#define _symat( inp, pos ) ((inp)->sym[pos])
#define symat( inp, pos ) ((pos) > maxpos(inp) ? EOFSYM : (inp)->sym[pos])
/* the symbol at the given position */

#define attat( inp, pos ) ((inp)->att[pos])
/* the attribute of the symbol at the given position */

#define charat( inp, pos ) ((inp)->src[pos])
/* the character at the given position */

/* Die folgenden 3 Makros sind fuer strings im Text,
   d.i. identifier, keyword, multichar-operator, zu gebrauchen,
   NICHT fuer String-Literale ! */
#define nstrat( inp, pos ) ((inp)->att[pos].a),(&((inp)->src[pos]))
/* the string at the given position,
   represents the string as 2 parameters: length and char pointer.
   can be e.g. used in printf using the format "%.*s" */
/* parts: */
#define strlenat( inp, pos ) ((inp)->att[pos].a)
#define strptrat( inp, pos ) (&((inp)->src[pos]))


// String-Literal: liefert STRING* (Ausgabe einfacher mit prsymat())
#define STRINGat( inp, pos ) ((inp)->att[pos].ps)
// string wie im Text mit Laenge und Pointer (fuer %.*s):
#define textstringat( inp, pos ) nstring(STRINGat(inp,pos)->txt)
// "wirklicher" string mit Laenge und Pointer:
#define string_at( inp, pos ) nstring(STRINGat(inp,pos)->val)
// "wirklicher" string pointer/Laenge:
#define string_ptr_at( inp, pos ) ((STRINGat(inp,pos)->val).str)
#define string_len_at( inp, pos ) ((STRINGat(inp,pos)->val).len)


// Zahl: liefert NUMBER* (Ausgabe einfacher mit prsymat())
#define NUMBERat( inp, pos ) ((inp)->att[pos].pn)


/* functions that support automatic scan: */
void scan_initauto( SCAN * inp,
   char *   comment_begin,       // start of comment block (nested)
   char *   comment_end,         // end of comment block
   char *   comment_line,        // start of comment until end of line
   char *   charsym,             // 1-char symbols
   _KEY_ *  optab,               // multi-char operators            -- table must end w. 0-entry
   _KEY_ *  keytab,              // keywords, alphabetically sorted -- table must end w. 0-entry
   _SYMBOL_ charlit_sym,         // if != 0: automatic char literal handling
   char     charlit_begin_end,   // e.g. '  (use 1 if charlit_sym == 0 !)
   _SYMBOL_ stringlit_sym,       // if != 0: automatic string literal handling
   char     stringlit_begin_end, // e.g. "
   int      feature_flags,       // see feature_... below
   int      esc_flags,           // flags for char and string literals // escape_... below
   _SYMBOL_ number_sym           // if != 0: automatic number  handling
  );
/* prepares automatic scanning, tells generated tables to the scanner */

// feature flags
#define feature_comment_not_nested 0x0001  // like in C
#define feature_case_sensitive     0x0002  // like in C
#define feature_hex_numbers        0x0004  // like in C, 0x...
#define feature_octal_numbers      0x0008  // like in C, 0...

// flags for lit_char_c and lit_string_c:
#define escape_double_start_char   0x0001  // like in Pascal
#define escape_a_allowed           0x0002
#define escape_b_allowed           0x0004
#define escape_f_allowed           0x0008
#define escape_n_allowed           0x0010
#define escape_r_allowed           0x0020
#define escape_t_allowed           0x0040
#define escape_v_allowed           0x0080
#define escape_x_allowed           0x0100  // erlaubt auch 0xAB-Zahlen
#define escape_X_allowed           0x0200
#define escape_octal_allowed       0x0400  // \012
#define escape_hashdec_allowed     0x1000  // #27


#define escape_std_c (escape_a_allowed | escape_b_allowed | escape_f_allowed |   \
                      escape_n_allowed | escape_r_allowed | escape_t_allowed |   \
                      escape_v_allowed | escape_x_allowed | escape_octal_allowed )

#define escape_turbo_pascal (escape_double_start_char | escape_hashdec_allowed)


int scan_auto( SCAN * inp, int pos );
/* scans using the tables given in scan_inittables(), including comment !
   returns at each non-trivial symbol -
   there no action is required - except for user symbols.
   scan_auto() returns the next position to look at, i.e. after the just
   scanned symbol. For a user symbol, i.e. a symbol, that cannot be handlled automatically,
   scan_auto() returns the same position that it got as input.
   You must handle this case, otherwise you will end in an endless loop...
   Note that words (like Identifiers) cause calling the user function wordsymbol().
   Call scan_auto() in a loop !
*/

int scan_comment( SCAN * inp, int pos );
/* returns 1st position (starting at pos) outside a comment */


/* user symbol helper functions: */
int lit_hexint( SCAN * inp, int pos );
/* scans hexadecimal number, integer, start at digit, returns pos after */

void scan_dump( FILE * f, SCAN * inp );
// fuer Testzwecke

#endif /* _SCAN_H */

