/*
   xbnfgen version 0.8  Copyright (C) 2007-2016 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/

/* xbnfgen.c
   simple parser generator from xbnf
   see xbnf.xbnf and scan.h
*/

#define VERSION "0.8"

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "port.h"
#include "xbnfscan.h"
#include "salloc.h"
#include "set2int.h"

#define false 0
#define true  1

/* the symbols are stored in a binary tree,
   NODEDATA is information we need per node
   -- poor man's inheritance */
/*
   symvalue: 0 for nonterminals, symbol value for terminals
   first:    FIRST(nonterminal), not used for terminals
*/
#define NODEDATA int symvalue; _SET first;
#include "btree.h"

/*
   phase 1    determine First for all nonterminals
   phase 2... generate code ...
*/
static int phase;
#define FIRST    1
#define DECLARE  2
#define INIT     3
#define CODE     4

/* flag bits used in the flags parameter: */
#define F_GENCODE 1


static NODE * symbols;
static int firstchanged;

static _SET sequence_first;
static _SET simple_first;


/* We keep names for the terminal symbols in the array termname[].
   The symbol values start at TERMOFFSET, newterm is the first unused
   value */
#define TERMOFFSET 129
static char * termname[CARDINAL-TERMOFFSET];
static int newterm;

static int level;      /* for indentation */
static int pass;       /* just counting how often language() was called */

static _SET singlecharsymbols;

#define INDENT 3

// output files
static FILE * parsout;
static FILE * symout;
static char * parsoutname;
static char * symoutname;

// language to be generated
#define C      1
#define PASCAL 2
static int g_lang = 0;
static int setnomax = -1;

static int nkeywordterminal;
static int noperatorterminal;


typedef struct
{
   char * symbol;
   char * name;
   int symbollen;
   int namelen;
} SYMBOLNAME;

typedef struct
{
   char chrsymbol;
   char * name;
} CHARNAME;

static CHARNAME charname[] = // used for combined names
{
   { ' ' , "BLANK_"        },
   { '!' , "EXCLAM_"       },
   { '"' , "QUOTE_"        },
   { '#' , "HASH_"         },
   { '$' , "DOLLAR_"       },
   { '%' , "PERCENT_"      },
   { '&' , "AND_"          },
   { '\'', "APOSTROPH_"    },
   { '(' , "OPENPAR_"      },
   { ')' , "CLOSEPAR_"     },
   { '*' , "ASTERISK_"     },
   { '+' , "PLUS_"         },
   { ',' , "COMMA_"        },
   { '-' , "MINUS_"        },
   { '.' , "DOT_"          },
   { '/' , "SLASH_"        },
   { ':' , "COLON_"        },
   { ';' , "SEMICOLON_"    },
   { '<' , "LT_"           },
   { '=' , "EQ_"           },
   { '>' , "GT_"           },
   { '?' , "QUESTION_"     },
   { '@' , "AT_"           },
   { '[' , "OPENBRACKET_"  },
   { '\\', "BACKSLASH_"    },
   { ']' , "CLOSEBRACKET_" },
   { '^' , "CARET_"        },
   { '_' , "UNDERBAR_"     },
   { '`' , "BACKTICK_"     },
   { '{' , "OPENBRACE_"    },
   { '|' , "BAR_"          },
   { '}' , "CLOSEBRACE_"   },
   { '~' , "TILDE_"        },
};

static SYMBOLNAME symbolname[] =
{
   { ":=", "ASSIGN_" },
   { "->", "ARROW_"  },
   { "<=", "LEQ_"    },
   { ">=", "GEQ_"    },
};


int iprintf( int ilevel, char * format, ... )
/* ================================================================ */
/* like printf  but with indentation */
{
int rc;
va_list ap;

   fprintf( parsout, "%*s", ilevel * INDENT, "" );
   va_start( ap, format );
   rc = vfprintf( parsout, format, ap );
   va_end( ap );
   return rc;
}


static char * basename( char * filename )
/* ---------------------------------------------------------------- */
{
   char * begin;
   char * end;

   if( !filename )
      return "";

   // Beginn des Basenames ist nach dem letzten '/' oder am Anfang
   begin = strrchr( filename, DIRSEP );
   if( begin )
      begin++;
   else
      begin = filename;

   // Ende des Basenamens ist der letzte '.' oder das Ende
   end = strrchr( begin, '.' );
   if( !end )
      end = begin + strlen(begin);

   // wir aendern im Original ...
   *end = 0;
   return begin;
}


static int iscname( int len, char * nam )
/* ---------------------------------------------------------------- */
{
int i;
char ch;

   for( i = 0; i < len; i++ )
   {
      ch = nam[i];
      if( !(isalnum(ch) || (ch == '_')) )
      {
         return 0; /* not a C name */
      }
   }
   return 1;
}


static char * getcharname( int c )
/* ---------------------------------------------------------------- */
{
   int i;
   char * name = 0;

   for( i = 0; i < arraysize(charname); i++ )
   {
      if( c == charname[i].chrsymbol )
      {
         name = charname[i].name;
         break;
      }
   }
   return name;
}


static char * getcharnamemust( int c )
/* ---------------------------------------------------------------- */
{
   static char buf[4];

   char * name = getcharname(c);
   if( name )
      return name;

   buf[0] = '\'';
   buf[1] = c;
   buf[2] = '\'';
   buf[3] = 0;
   return buf;
}


static char * guessname( int slen, char * str, int * namelen  )
/* ---------------------------------------------------------------- */
{
   int i, k, d;
   char c;
   char * ptr;
   char * cname;
   static char namebuf[200];

   // multi char name ?
   for( i = 0; i < arraysize(symbolname); i++ )
   {
      if( symbolname[i].symbollen == 0 )
         symbolname[i].symbollen = strlen( symbolname[i].symbol );

      if( symbolname[i].namelen == 0 )
         symbolname[i].namelen = strlen( symbolname[i].name );

      if( nstrcmp( slen, str, symbolname[i].symbollen, symbolname[i].symbol ) == 0 )
      {
         *namelen = symbolname[i].namelen;
         return symbolname[i].name;
      }
   }
   // try to build from single char names
   d = 0;
   for( k = 0; k < slen; k++ )
   {
      if( d > (sizeof(namebuf) - 20) )
         return 0;      // would become too long

      c = str[k];

      // look in charname
      cname = getcharname( c );
      if( cname )
      {
         strcpy( namebuf+d, cname );
         d += strlen(cname);
      }
      else
      {
         d = 0;
         break;
      }
   }

   if( d > 0 )
   {
      *namelen = d;
      namebuf[d] = 0;
      ptr = salloc( d + 1 );
      strcpy( ptr, namebuf );
      return ptr;
   }

   // give up
   return 0;
}


static int termsym( SCAN * inp, int pos )
/* ---------------------------------------------------------------- */
/* returns the symbol value for the given terminal of the
   destination language, returns -1 in case of an error

   The string must be used as C name afterwards !
*/
{
NODE * term;
int isnew;
int namelen;
char * name;

   /* 1st attempt: it is already a valid C name ... */
   if( iscname( string_at( inp, pos ) ) )
   {
      namelen = string_len_at(inp,pos);
      name    = string_ptr_at(inp,pos);
   }
   /* 2nd attempt: we've got some predefined names... */
   else if( (name = guessname( string_at( inp, pos ), &namelen )) != 0 )
   {
      // ok, nothing to do here
   }
   else
   {
      MERROR( "cannot guess a symbol value name for", strf( "%.*s", textstringat(inp,pos) ) );
      return -1;
   }

   /* we put the terminals into the same tree... */
   term = bt_insert( symbols, string_at( inp, pos ), &isnew );
   if( isnew )
   {
      newterm++;
      if( newterm >= CARDINAL )
         MFATAL_POS( pos, "too many terminal symbols", strf("max %d",CARDINAL-TERMOFFSET) );
      term->symvalue = newterm;
      /* just use the terminal name in UPPERCASE */
      termname[newterm-TERMOFFSET] = strupr( strfalloc( "%.*s%s", namelen, name, g_lang == PASCAL ? "_" : "" ) );
   }

   return term->symvalue;
}


char * prset( _SET * set, int printname )
/* ---------------------------------------------------------------- */
/* prints a symbol set to a static buffer, printname = flag, print either '.' (false) or DOT_ (true) */
{
static char buf[10*CARDINAL];
int i;
char * s;
char * cname;

   s = buf;
   for( i = 0; i < TERMOFFSET; i++ )
   {
      if( set_in( set, i ) )
      {
         *s++ = ' ';
         cname = getcharname(i);
         if( printname && cname )
         {
            strcpy( s, cname );
            s += strlen( s );
         }
         else
         {
            *s++ = '\'';
            if( i < ' ' )
            {
               *s++ = '\\';  /*octal */
               *s++ = '0';
               *s++ = '0' + i / 8;
               *s++ = '0' + i % 8;
            }
            else if( i == '\'' )
            {
               *s++ = '\\';
               *s++ = i;
            }
            else
            {
               *s++ = i;
            }
            *s++ = '\'';
         }
      }
   }
   for( i = TERMOFFSET; i < CARDINAL; i++ )
   {
      if( set_in( set, i ) )
      {
         *s++ = ' ';
         strcpy( s, termname[i-TERMOFFSET] );
         s += strlen( s );
      }
   }
   *s = 0;
   return buf;
}


static void printfirstnode( NODE * node )
/* ---------------------------------------------------------------- */
{
   if( node->symvalue == 0 ) /* nonterminals */
      fprintf( parsout, "first[ %-14.*s ] =%s\n", nstring(node->name), prset( &(node->first), false ) );
}


static int isletter( int ch )
/* ---------------------------------------------------------------- */
{
   return ((ch >= 'A') && (ch <= 'Z')) || ((ch >= 'a') && (ch <= 'z')) || (ch == '_');
}


static void printkeywordterminal( NODE * node )
/* ---------------------------------------------------------------- */
{
   if( (node->symvalue != 0) && isletter(node->name.str[0]) && isupper(node->name.str[1]) ) /* terminal, key word */
   {
      if( g_lang == C )
         fprintf( symout, "{ \"%s\" %*s , %-20s },\n", strlwr( strfalloc( "%.*s", nstring(node->name) ) ), 20-node->name.len, "", termname[node->symvalue-TERMOFFSET] );
      else if( g_lang == PASCAL )
         fprintf( symout, "  ( word: '%s' %*s ; sym: %-20s ),\n", strlwr( strfalloc( "%.*s", nstring(node->name) ) ), 20-node->name.len, "", termname[node->symvalue-TERMOFFSET] );
   }
}


static void countkeywordterminal( NODE * node )
/* ---------------------------------------------------------------- */
{
   if( (node->symvalue != 0) && isletter(node->name.str[0]) && isupper(node->name.str[1]) ) /* terminal, key word */
      nkeywordterminal++;
}


static void printscannerterminal( NODE * node )
/* ---------------------------------------------------------------- */
{
   if( (node->symvalue != 0) && isletter(node->name.str[0]) && islower(node->name.str[1]) ) /* Upper lower : scanner symbol */
      fprintf( symout, "   %s,\n", termname[node->symvalue-TERMOFFSET] );
}


static void printoperatorterminal( NODE * node )
/* ---------------------------------------------------------------- */
{
   if( (node->symvalue != 0) && !isletter(node->name.str[0]) ) /* terminal, operator */
   {
      if( g_lang == C )
         fprintf( symout, "{ \"%.*s\" %*s , %-20s },\n", nstring(node->name), 20-node->name.len, "", termname[node->symvalue-TERMOFFSET] );
      else if( g_lang == PASCAL )
         fprintf( symout, "  ( word: '%.*s' %*s ; sym: %-20s ),\n", nstring(node->name), 20-node->name.len, "", termname[node->symvalue-TERMOFFSET] );
   }
}


static void countoperatorterminal( NODE * node )
/* ---------------------------------------------------------------- */
{
   if( (node->symvalue != 0) && !isletter(node->name.str[0]) ) /* terminal, operator */
      noperatorterminal++;
}


static int genfirstset( _SET * first )
/* ---------------------------------------------------------------- */
/* generates a set named first_# having all the elements of first
   returns the number (#)                                        */
{
int setno, isnew, i;
/* need to count the unique sets for the INIT phase, because nothing
   is new, then */
static int init_cnt = 0;

   /* we don't want to generate what we already have. therefore
      we use set2int to generate unique sets */
   setno = set2int( first, &isnew );
   if( g_lang == C )
   {
      if( phase == DECLARE )
      {
         if( isnew )
            fprintf( parsout, "static _SET first_%d;\n" , setno );
      }
      else if( phase == INIT )
      {
         if( setno > init_cnt )
         {
            init_cnt = setno;
            fprintf( parsout, "\n" );
            iprintf( 1, "set_empty( &first_%d );\n", setno );
            for( i = 0; i < TERMOFFSET; i++ )
               if( set_in( first, i ) )
                  iprintf( 1, "set_put( &first_%d, '%s%c' );\n", setno, i == '\'' ? "\\" : "", i );
            for( i = TERMOFFSET; i < CARDINAL; i++ )
               if( set_in( first, i ) )
                  iprintf( 1, "set_put( &first_%d, %s );\n", setno, termname[i-TERMOFFSET]  );
         }
      }
   }
   else if( g_lang == PASCAL )
   {
      if( phase == DECLARE )
      {
         if( isnew )
         {
            if( setno > setnomax )
               setnomax = setno;
         }
      }
      else if( phase == INIT )
      {
         if( setno > init_cnt )
         {
            char * comma = "";
            init_cnt = setno;
            fprintf( parsout, "   first_%d := [", setno );
            for( i = 0; i < TERMOFFSET; i++ )
               if( set_in( first, i ) )
               {
                  fprintf( parsout, "%s\n      %s", comma, getcharnamemust(i) );
                  comma = ",";
               }
            for( i = TERMOFFSET; i < CARDINAL; i++ )
               if( set_in( first, i ) )
               {
                  fprintf( parsout, "%s\n      %s", comma, termname[i-TERMOFFSET]  );
                  comma = ",";
               }
            fprintf( parsout, "\n   ];\n" );
         }
      }
   }
   return setno;
}


static char * pascalstmt2ndhalf( char * stmt )
{
   if( strcmp(stmt,"if") == 0 )
      return "then";
   if( strcmp(stmt,"while") == 0 )
      return "do";
   return "";
}


static void gendecision( char * stmt, _SET * first )
/* ---------------------------------------------------------------- */
/* generates a if/while until { using the symbols in first
   stmt = "if" or "while"                                 */
{
int num, n;

   num = set_power( first ); /* number of symbols in First */
   if( num == 1 )
   {
      if( phase == CODE )
      {
         if( g_lang == C )
         {
            iprintf( level, "%s( symat(inp,pos) ==%s )\n", stmt, prset(first, false) );
            iprintf( level, "{\n" );
         }
         else if( g_lang == PASCAL )
         {
            iprintf( level, "%s symat(inp,pos) =%s %s\n", stmt, prset(first, true), pascalstmt2ndhalf(stmt) );
            iprintf( level, "begin\n" );
         }
      }
   }
   else
   {
      n = genfirstset( first );
      if( phase == CODE )
      {
         if( g_lang == C )
         {
            iprintf( level, "%s( set_in( &first_%d, symat(inp,pos) ) )\n", stmt, n );
            iprintf( level, "{\n" );
         }
         else if( g_lang == PASCAL )
         {
            iprintf( level, "%s symat(inp,pos) in first_%d %s\n", stmt, n, pascalstmt2ndhalf(stmt) );
            iprintf( level, "begin\n" );
         }
      }
   }
}


/* XBNF nonterminals begin ======================================== */

static void language( SCAN * inp, int pos );
static int nonterminal( SCAN * inp, int pos );
static int sequence( SCAN * inp, int pos, _SET * first, int flags );
static int simpleorany( SCAN * inp, int pos, _SET * first, int flags );
static int simple( SCAN * inp, int pos, _SET * first, int flags );
static int finitany( SCAN * inp, int pos, _SET * first, int flags );
static int group( SCAN * inp, int pos, _SET * first, int flags );
static int option( SCAN * inp, int pos, _SET * first, int flags );
static int iteration( SCAN * inp, int pos, _SET * first, int flags );
static int alternative( SCAN * inp, int pos, _SET * first, int flags );


static void language( SCAN * inp, int pos )
/* ---------------------------------------------------------------- */
{
   pass++;
   while( symat( inp, pos ) == LOWID )
   {
      pos = nonterminal( inp, pos );
   }
   if( (g_lang == PASCAL) && (phase == CODE) )
   {
      iprintf( 0, "\nbegin\n" );
      iprintf( 1, "init;\n" );
      iprintf( 0, "end.\n" );
   }
}


static int nonterminal( SCAN * inp, int pos )
/* ---------------------------------------------------------------- */
{
int name;
_SET first;
NODE * thissym;
int isnew;

   set_empty( &first );
   level = 1;
   name = pos;  /* where the name of the nonterminal begins */
   pos = eatsym( inp, pos, LOWID );
   pos = eatsym( inp, pos, EQ );
   thissym = bt_insert( symbols, nstrat( inp, name ), &isnew );

   if( g_lang == C )
   {
      if( phase == DECLARE )
      {
         static int startsymbol = 1; // we assume here that the 1st non-terminal is the start symbol
         fprintf( parsout, "%s int _%.*s( SCAN * inp, int pos );\n", startsymbol ? "      " : "static", nstrat( inp, name ) );
         startsymbol = 0;
      }
      else if( phase == CODE )
      {
         static int startsymbol = 1; // we assume here that the 1st non-terminal is the start symbol
         fprintf( parsout, "\n%s int _%.*s( SCAN * inp, int pos )\n/* ---------------------------------------------------------------- */\n{\n",
            startsymbol ? "" : "static", nstrat( inp, name ) );
         iprintf( 1, "_PRINT_(( \"%%*s%.*s-(%%s)\\n\", level*INDENT, \"\", prsymat(inp,pos) ));\n", nstrat( inp, name ) );
         iprintf( 1, "level++;\n" );
         startsymbol = 0;
      }
   }
   else if( g_lang == PASCAL )
   {
      if( phase == DECLARE )
      {
         fprintf( parsout, "function _%.*s( var inp: SCAN; pos: integer ) : integer;\n", nstrat( inp, name ) );
      }
      else if( phase == CODE )
      {
         fprintf( parsout, "\nfunction _%.*s( var inp: SCAN; pos: integer ) : integer;\n(* ---------------------------------------------------------------- *)\nbegin\n",
                                         nstrat( inp, name ) );
         iprintf( 1, "_PRINT_( \'%%*s%.*s-(%%s)\', [ level*INDENT, \'\', prsymat(inp,pos) ] );\n", nstrat( inp, name ) );
         iprintf( 1, "level := level+1;\n" );
      }
   }

   /* in pass 1 everything must be new ! */
   if( !isnew && (pass == 1) )
      MERROR_POS( name, "nonterminal defined twice", prsymat(inp,name) );

   pos = abs(sequence( inp, pos, &first, F_GENCODE ));
   pos = eatsym( inp, pos, DOT );

   if( isnew )
   {
      thissym->first = first;     /* can happen only in phase FIRST */
      firstchanged = 1;
   }
   else
   {
      /* collect First() : */
      set_union( &first, &(thissym->first), &first );
      if( (phase == FIRST) && !set_eq( &first, &(thissym->first) ) )
      {
         thissym->first = first;  /* can happen only in phase FIRST */
         firstchanged = 1;        /*  even w/o the test in if() !   */
      }
   }

   if( phase == CODE )
   {
      if( g_lang == C )
      {
         iprintf( 1, "level--;\n" );
         iprintf( 1, "return pos;\n}\n\n" );
      }
      else if( g_lang == PASCAL )
      {
         iprintf( 1, "level := level-1;\n" );
         iprintf( 1, "_%.*s := pos;\nend;\n\n", nstrat( inp, name ) );
      }
   }
   return pos;
}


static int sequence( SCAN * inp, int pos, _SET * first, int flags )
/* ---------------------------------------------------------------- */
/* opaque for First in a sequence -> returns positive position
   not opaque (i.e. First not complete) - returns neg. position */
{
_SET morefirst;
char transfirst; /* flag */

   pos = simple( inp, pos, first, flags );
   transfirst = pos < 0;
   pos = abs(pos);
   while( set_in( &sequence_first, symat( inp, pos ) ) )
   {
      set_empty( &morefirst );
      pos = simpleorany( inp, pos, &morefirst, flags );
      if( transfirst )
         set_union( first, first, &morefirst );
      transfirst = transfirst && (pos < 0); /* transparency stops at 1st opaque simnple */
      pos = abs(pos);
   }
   if( transfirst )
      pos = -pos;
   return pos;
}


static char * assignsym( void )
{
   if( g_lang == PASCAL )
      return ":=";
   else
      return "=";
}


static char * beginsym( void )
{
   if( g_lang == PASCAL )
      return "begin";
   else
      return "{";
}


static char * endsym( void )
{
   if( g_lang == PASCAL )
      return "end;";
   else
      return "}";
}


static char * caseendsym( void )
{
   if( g_lang == PASCAL )
      return "end;";
   else
      return "break;";
}


static int simpleorany( SCAN * inp, int pos, _SET * first, int flags )
/* ---------------------------------------------------------------- */
{
   switch( symat(inp,pos) )
   {
   case LPAREN:
   case LBRACK:
   case LBRACE:
   case LT:
   case LOWID:
   case TERMINAL:
      pos = simple( inp, pos, first, flags );
      break;
   case MINUS:
      pos = finitany( inp, pos, first, flags );
      break;
   default:
      MERROR( "expected one of", prtermsymset( inp, &sequence_first ) );
      pos = nextsym( inp, pos );
   }
   return pos;
}


static int processterminal( SCAN * inp, int pos, _SET * symset )
/* packt das Terminalsymbol der Zielsprache in symset.
   returns 1 if ok, 0 if error */
{
int sy;

   if( (string_len_at(inp,pos) == 1) && (charat(inp,pos) > ' ') && (charat(inp,pos) < 127) )
   {
      /* single char symbol - use the code as value: */
      set_put( symset, charat(inp,pos) );

      set_put( &singlecharsymbols, charat(inp,pos) );
   }
   else
   {
      /* Here we must invent a symbol value and name for the terminal ! */
      sy = termsym( inp, pos );
      if( sy >= 0 )
         set_put( symset, sy );
      else
         return 0; /* error */
   }
   return 1;
}


static int simple( SCAN * inp, int pos, _SET * first, int flags )
/* ---------------------------------------------------------------- */
/* opaque for First in a sequence -> returns positive position
   not opaque (i.e. First not complete) - returns neg. position */
{
NODE * thissym;
int comp, npos;
int n = 0;

   switch( symat( inp, pos ) )
   {
   case LOWID:
      thissym = bt_search( symbols, nstrat( inp, pos ), &comp );
      if( comp == 0 )
      {            /* existing nonterm */
         *first = thissym->first;

         if( (flags & F_GENCODE) && (phase == CODE) )
            iprintf( level, "pos %s _%.*s( inp, pos );\n", assignsym(), nstrat(inp,pos) );
      }
      else if( phase == CODE ) /* now we should know everything */
         MERROR( "unknown nonterminal", 0 );

      pos = nextsym( inp, pos ); /* opaque */
      break;

   case TERMINAL:
      if( !processterminal(inp,pos,first) )
      {
         pos = nextsym( inp, pos ); /* error */
         break;
      }

      if( (flags & F_GENCODE) && (phase == CODE) )
      {
         /* if first contains only 1 element we can just eat that.
            otherwise we skip the symbol and hope it was ok.
            The 2nd case happens in case statements, where indead
            the check happens earlier, so skipping would be ok. */
         if( set_power(first) == 1 )
            iprintf( level, "pos %s eatsym( inp, pos,%s );\n", assignsym(), prset(first, g_lang == PASCAL) );
         else
            iprintf( level, "pos %s nextsym( inp, pos ); /*%s */\n", assignsym(), prset(first, false) );
      }
      pos = nextsym( inp, pos ); /* opaque */
      break;

   case LPAREN:
      pos = group( inp, pos, first, flags ); /* opaque ? -> same as group */
      break;

   case LBRACK:
      pos = - option( inp, pos, first, flags );    /* not opaque ! */
      break;

   case LBRACE:
      pos = - iteration( inp, pos, first, flags ); /* not opaque ! */
      break;

   case LT:
      pos = nextsym( inp, pos );

      if( (flags & F_GENCODE) && (phase == CODE) )
      {
         if( g_lang == C )
         {
            iprintf( level, "switch( symat(inp,pos) )\n" );
            iprintf( level, "{\n" );
         }
         else if( g_lang == PASCAL )
         {
            iprintf( level, "case symat(inp,pos) of\n" );
         }
      }
      /* all the First(variant) must be pairwise disjoint, otherwise
         this switch approach cannot work. ...
         see alternative() */

      npos = alternative( inp, pos, first, flags & ~F_GENCODE );

      if( (flags & F_GENCODE) && (phase > FIRST) )
         n = genfirstset( first );

      npos = alternative( inp, pos, first, flags );

      if( (flags & F_GENCODE) && (phase == CODE) )
      {
         if( g_lang == C )
         {
            iprintf( level, "default:\n" );
            iprintf( level+1, "MERROR( \"expected one of\", prtermsymset( inp, &first_%d ) );\n", n );
            iprintf( level+1, "pos = nextsym( inp, pos );\n" );
            iprintf( level, "}\n" );
         }
         else if( g_lang == PASCAL )
         {
            iprintf( level, "else\n" );
            iprintf( level+1, "begin\n" );
            iprintf( level+2, "MERROR( inp, pos, \'expected one of\', prtermsymset( inp, first_%d ) );\n", n );
            iprintf( level+2, "pos := nextsym( inp, pos )\n" );
            iprintf( level+1, "end\n" );
            iprintf( level, "end;\n" );
         }
      }

      pos = npos;
      pos = eatsym( inp, pos, GT ); /* opaque */
      break;

   default:
      MERROR( "expected one of", prtermsymset( inp, &simple_first ) );
      pos = nextsym( inp, pos );  // hier sollten wir eher springen bis ...
      break;
   }
   return pos;
}


static int finitany( SCAN * inp, int pos, _SET * first, int flags )
/* ---------------------------------------------------------------- */
{
   _SET s;
   set_empty( &s );

   pos = eatsym( inp, pos, '-' );

   if( !processterminal(inp,pos,&s) )
   {
      return nextsym( inp, pos ); /* error */
   }

   if( (flags & F_GENCODE) && (phase == CODE) )
   {
      if( g_lang == C )
      {
         iprintf( level, "while( symat(inp,pos) !=%s )\n", prset(&s,0) );
         iprintf( level, "{\n" );
         iprintf( level+1, "pos = nextsym( inp, pos );\n" );
         iprintf( level, "}\n" );
         iprintf( level, "pos = nextsym( inp, pos );\n" );
      }
      else if( g_lang == PASCAL )
      {
         iprintf( level, "while symat(inp,pos) <>%s do\n", prset(&s,1) );
         iprintf( level, "begin\n" );
         iprintf( level+1, "pos := nextsym( inp, pos );\n" );
         iprintf( level, "end;\n" );
         iprintf( level, "pos := nextsym( inp, pos );\n" );
      }
   }
   pos = eatsym( inp, pos, TERMINAL );
   return pos;
}


static int group( SCAN * inp, int pos, _SET * first, int flags )
/* ---------------------------------------------------------------- */
/* opaque for First in a sequence -> returns positive position
   not opaque (i.e. First not complete) - returns neg. position */
{
char transfirst; /* flag */

   pos = eatsym( inp, pos, LPAREN );

   if( (flags & F_GENCODE) && (phase == CODE) )
      iprintf( level, "%s\n", beginsym() );

   level++;
   pos = sequence( inp, pos, first, flags );
   transfirst = pos < 0;
   pos = abs(pos);
   level--;

   if( (flags & F_GENCODE) && (phase == CODE) )
      iprintf( level, "%s\n", endsym() );

   pos = eatsym( inp, pos, RPAREN );
   if( transfirst )
      pos = -pos;
   return pos;
}


static int option( SCAN * inp, int pos, _SET * first, int flags )
/* ---------------------------------------------------------------- */
{
int npos;

   pos = eatsym( inp, pos, LBRACK );

   /* Here is a problem:  (same in simple/iteration/alternative)
      We need to call sequence to generate the code it represents.
      But before we do that we need to know First(of this sequence)
      (to generate the if statement that precedes the sequence-code)
      for which we need to call sequence...
      Solution: call sequence twice with an extra flag F_GENCODE
      to indicate the call in which code can ge generated.  */

   set_empty( first );
   npos = sequence( inp, pos, first, flags & ~F_GENCODE );
   /* now first should be First(this sequence) ! */

   if( (flags & F_GENCODE) && (phase > FIRST) )
      gendecision( "if", first );

   level++;
   npos = sequence( inp, pos, first, flags );
   level--;

   if( (flags & F_GENCODE) && (phase == CODE) )
      iprintf( level, "%s\n", endsym() );

   pos = abs(npos);
   pos = eatsym( inp, pos, RBRACK );
   return pos;
}


static int iteration( SCAN * inp, int pos, _SET * first, int flags )
/* ---------------------------------------------------------------- */
{
int npos;

   pos = eatsym( inp, pos, LBRACE );

   set_empty( first );
   npos = sequence( inp, pos, first, flags & ~F_GENCODE );

   if( (flags & F_GENCODE) && (phase > FIRST) )
      gendecision( "while", first );

   level++;
   npos = sequence( inp, pos, first, flags );
   level--;

   if( (flags & F_GENCODE) && (phase == CODE) )
      iprintf( level, "%s\n", endsym() );

   pos = abs(npos);
   pos = eatsym( inp, pos, RBRACE );
   return pos;
}


static int alternative( SCAN * inp, int pos, _SET * first, int flags )
/* ---------------------------------------------------------------- */
{
_SET morefirst;
_SET check;
int i, npos;

   if( symat( inp, pos ) == LBRACK )
      MERROR( "Option directly as alternative does not make sense.", "What do you mean ?" );

   set_empty( first );
   npos = sequence( inp, pos, first, flags & ~F_GENCODE );

   if( (flags & F_GENCODE) && (phase == CODE) )
   {
      /* make a case for each symbol in first: */
      if( g_lang == C )
      {
         for( i = 0; i < TERMOFFSET; i++ )
            if( set_in( first, i ) )
               iprintf( level, "case '%s%c':\n", i == '\'' ? "\\" : "", i );

         for( i = TERMOFFSET; i < CARDINAL; i++ )
            if( set_in( first, i ) )
               iprintf( level, "case %s:\n", termname[i-TERMOFFSET] );
      }
      else if( g_lang == PASCAL )
      {
         char * comma = "";
         iprintf( level, "" );

         for( i = 0; i < TERMOFFSET; i++ )
            if( set_in( first, i ) )
            {
               fprintf( parsout, "%s%s", comma, getcharnamemust(i) );
               comma = ", ";
            }

         for( i = TERMOFFSET; i < CARDINAL; i++ )
            if( set_in( first, i ) )
            {
               fprintf( parsout, "%s%s", comma, termname[i-TERMOFFSET] );
               comma = ", ";
            }

         fprintf( parsout, ":\n" );
      }
   }

   if( (flags & F_GENCODE) && (phase == CODE) && (g_lang == PASCAL) )
      iprintf( level+1, "begin\n" );

   level = level + 1 + (g_lang == PASCAL);
   npos = sequence( inp, pos, first, flags );
   level = level - 1 - (g_lang == PASCAL);

   if( (flags & F_GENCODE) && (phase == CODE) )
      iprintf( level+1, "%s\n", caseendsym() );

   pos = abs(npos);

   if( symat( inp, pos ) == BAR )
   {
      pos = nextsym( inp, pos );
      set_empty( &morefirst );
      pos = alternative( inp, pos, &morefirst, flags );

      /* the new set morefirst must be disjoint with the
         already known first: */
      set_intersect( &check, first, &morefirst );
      if( set_power( &check ) != 0 )
         MERROR( "alternatives are not disjoint. overlapping:", prset(&check, false) );

      /* collect First symbols: */
      set_union( first, first, &morefirst );
   }
   return pos;
}

/* XBNF nonterminals end   ======================================== */


/*------------------------------------------------------------------*/
static FILE * outopen( char * filename )
{
   FILE * f;

   f = fopen( filename, "wt" );
   if( !f )
      MFATAL( "Cannot create", filename );

   return f;
}


static void usage()
/* ---------------------------------------------------------------- */
{
   fprintf( stderr, "USAGE: xbnfgen SyntaxIn.xbnf [-Language] ParsertemplateOut SymbolsOut\n" );
   fprintf( stderr, "               Language: C (default) | Pascal\n" );
   exit( 1 );
}

static void setlanguage( char * lang )
/* ---------------------------------------------------------------- */
{
   if( stricmp( lang, "C" ) == 0 )
   {
      g_lang = C;
   }
   else if( stricmp( lang, "PASCAL" ) == 0 )
   {
      g_lang = PASCAL;
   }
   else
   {
      fprintf( stderr, "language %s is not supported\n", lang );
      exit( 5 );
   }
}


// code generation routines - using g_lang
static void mk_header( char * input )
/* ---------------------------------------------------------------- */
{
   if( g_lang == C )
   {
      fprintf( parsout, "/* generated by xbnfgen %s\n   input    = %s\n   language = C\n\n", VERSION, input );
      bt_foreach( symbols, symbols, printfirstnode );
      fprintf( parsout, "\n*/\n#define INDENT 3\n#define _PRINT_(X) printf X\nstatic int level;\n\n" );
   }
   else if( g_lang == PASCAL )
   {
      fprintf( parsout, "(* generated by xbnfgen %s\n   input    = %s\n   language = Pascal\n\n", VERSION, input );
      bt_foreach( symbols, symbols, printfirstnode );
      fprintf( parsout, "\n*)\nunit %s;\ninterface\nuses scanner;\n\n", basename(parsoutname) );
   }
}


static void mk_init( SCAN * inp )
/* ---------------------------------------------------------------- */
{
int i;

   if( g_lang == C )
   {
      fprintf( parsout, "\nstatic void init( void )\n/* ---------------------------------------------------------------- */\n{\n" );
      iprintf( 1, "level = 0;\n" );
      language( inp, nextsym(inp,0) );
      fprintf( parsout, "}\n\n" );
   }
   else if( g_lang == PASCAL )
   {
      fprintf( parsout, "\n\nimplementation\nuses msg, %s;\n\nconst indent = 3;\n\nvar\n   level: integer;\n", basename(symoutname) );

      for( i = 1; i <= setnomax; i++ )
         fprintf( parsout, "   first_%d: SYMBOLSET;\n" , i );

      fprintf( parsout, "\n\nprocedure init;\n(* ---------------------------------------------------------------- *)\nbegin\n" );
      iprintf( 1, "level := 0;\n" );
      language( inp, nextsym(inp,0) );
      fprintf( parsout, "end;\n\n" );
   }
}


static void mk_symbols( char * input )
/* ---------------------------------------------------------------- */
{
int i;

   if( g_lang == C )
   {
      fprintf( symout, "/* generated by xbnfgen %s\n   input    = %s\n   language = C\n*/\n\n", VERSION, input );
      fprintf( symout, "#ifndef   _SYMBOL_DEFINITIONS_\n#define   _SYMBOL_DEFINITIONS_\n\n" );

      // constants
      for( i = TERMOFFSET; i < CARDINAL; i++ )
         if( termname[i-TERMOFFSET][0] )
            fprintf( symout, "#define %-20s %3d\n", termname[i-TERMOFFSET], i );

      //  2 tables: keywords (starting w letter) and operators (other)
      fprintf( symout, "\n#ifdef    _SYMBOL_STRINGS_\n\n" );

      fprintf( symout, "static _KEY_ keytab[] = { /* alphabetically */\n" );
      bt_foreach( symbols, symbols, printkeywordterminal );
      fprintf( symout, "   {0,0,0}\n" );
      fprintf( symout, "};\n\n" );

      fprintf( symout, "static _SYMBOL_ termsym[] = {\n" );
      bt_foreach( symbols, symbols, printscannerterminal );
      fprintf( symout, "};\n\n" );

      fprintf( symout, "static _KEY_ optab[] = {\n" );
      bt_foreach( symbols, symbols, printoperatorterminal );
      fprintf( symout, "   {0,0,0}\n" );
      fprintf( symout, "};\n\n" );

      fprintf( symout, "static char charsym[] = {\n\"" );
      for( i = 0x21; i < 127; i++ )
         if( set_in( &singlecharsymbols, i ) )
            fprintf( symout, "%c", i );
      fprintf( symout, "\"\n};\n\n" );

      fprintf( symout, "#endif /* _SYMBOL_STRINGS_     */\n\n" );
      fprintf( symout, "#endif /* _SYMBOL_DEFINITIONS_ */\n" );
   }
   else if( g_lang == PASCAL )
   {
      fprintf( symout, "(* generated by xbnfgen %s\n   input    = %s\n   language = Pascal\n*)\n\n", VERSION, input );
      fprintf( symout, "unit %s;\ninterface\nuses scanner;\nconst\n", basename(symoutname) );

      // constants
      // bt_foreach( symbols, symbols, printscannerterminal );
      for( i = 0; i < TERMOFFSET; i++ )
         if( set_in( &singlecharsymbols, i ) )
            fprintf( symout, "   %-20s = %3d;\n", getcharnamemust(i), i );
      for( i = TERMOFFSET; i < CARDINAL; i++ )
         if( termname[i-TERMOFFSET][0] )
            fprintf( symout, "   %-20s = %3d;\n", termname[i-TERMOFFSET], i );

      //  2 tables: keywords (starting w letter) and operators (other)

      nkeywordterminal  = 0;
      bt_foreach( symbols, symbols, countkeywordterminal );
      if( nkeywordterminal > 0 )
      {
         fprintf( symout, "\n num_of_keywords = %d;\n", nkeywordterminal );
         fprintf( symout, " keytab: array [1..%d] of _KEY_ = ( (* alphabetically *)\n", nkeywordterminal+1 );
         bt_foreach( symbols, symbols, printkeywordterminal );
         fprintf( symout, "  ( word: 'zzdummy'               ; sym: 255                  )\n );\n\n" );
      }
      noperatorterminal = 0;
      bt_foreach( symbols, symbols, countoperatorterminal );
      if( noperatorterminal > 0 )
      {
         fprintf( symout, "\n num_of_multichar_operators = %d;\n", noperatorterminal );
         fprintf( symout, " optab: array [1..%d] of _KEY_ = (\n", noperatorterminal+1 );
         bt_foreach( symbols, symbols, printoperatorterminal );
         fprintf( symout, "  ( word: '_zzdummy'              ; sym: 255                  )\n );\n\n" );
      }

      fprintf( symout, "implementation\nend.\n" );
   }
}



int main( int argc, char * argv[] )
/* ---------------------------------------------------------------- */
{
SCAN * inp;
int i, ark;

   /* init */
   set_empty( &singlecharsymbols );

   set_empty( &sequence_first );
   set_put( &sequence_first, LOWID );
   set_put( &sequence_first, TERMINAL );
   set_put( &sequence_first, LPAREN );
   set_put( &sequence_first, LBRACK );
   set_put( &sequence_first, LBRACE );
   set_put( &sequence_first, LT );
   set_put( &sequence_first, MINUS );

   set_empty( &simple_first );
   set_put( &simple_first, LOWID );
   set_put( &simple_first, TERMINAL );
   set_put( &simple_first, LPAREN );
   set_put( &simple_first, LBRACK );
   set_put( &simple_first, LBRACE );
   set_put( &simple_first, LT );

   newterm = TERMOFFSET - 1;
   for( i = TERMOFFSET; i < CARDINAL; i++ )
      termname[i-TERMOFFSET] = "";
   pass = 0;

   symbols = bt_mktree( "m$", sizeof(NODE) );

   if( argc < 4 )
      usage();

   ark=1;
   inp = screadfile( argv[ark++] );
   if( inp )
      inp = scan( inp );

   if( argv[ark][0] == '-' )
      setlanguage( argv[ark++]+1 );
   else
      setlanguage( "c" );

   parsoutname = argv[ark++];
   symoutname  = argv[ark++];

   parsout = outopen( parsoutname );
   symout  = outopen( symoutname );

   if( inp )
   {
      fprintf( stderr, "xbnfgen %s (fr) 2016\n", VERSION );

      phase = FIRST; /* phase 1 : determine First(nonterminals) */
      do
      {
         firstchanged = 0;
         language( inp, nextsym(inp,0) );
      } while( firstchanged );

      /* now generate the code */
      mk_header( argv[1] );
      phase = DECLARE;
      language( inp, nextsym(inp,0) );

      phase = INIT;
      mk_init( inp );

      phase = CODE;
      language( inp, nextsym(inp,0) );

      /* generate symbol header file: */
      mk_symbols( argv[1] );
   }
   else
      MFATAL( "scan errors in", argv[1] );

   exit( 0 );
}


